package org.jboss.logging;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.LogManager;

import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import org.jboss.system.ServiceMBeanSupport;
import org.jboss.util.Strings;
import org.jboss.util.ThrowableListener;
import org.jboss.util.stream.Streams;

/**
 * Initializes the JDK logging framework.
 * 
 * <p>
 * Sets up a {@link ThrowableListener} to adapt unhandled throwables to a
 * logger.
 * 
 * <p>
 * 
 * @jmx:mbean name="jboss.system:type=JDKLoggerService,service=Logging"
 *            extends="org.jboss.system.ServiceMBean"
 * 
 * @version <tt>$Revision: 1.27 $</tt>
 */
public class JDKLoggerService extends ServiceMBeanSupport implements
		JDKLoggerServiceMBean {
	/**
	 * The default url for the configuration file. Reads the value from the
	 * system property <tt>org.jboss.logging.JDKLoggerService.configURL</tt>
	 * or if that is not set defaults to <tt>resource:loging.properties</tt>.
	 */
	public static final String DEFAULT_URL = System.getProperty(
			JDKLoggerService.class.getName() + ".configURL",
			"resource:logging.properties");

	/** The URL to the configuration file. */
	private URL configURL;

	/** The time in seconds between checking for new config. */
	private int refreshPeriod;

	/** The URL watch timer (in daemon mode). */
	private Timer timer = new Timer(true);

	/** The specialized timer task to watch our config file. */
	private URLWatchTimerTask timerTask;

	/**
	 * A flag to enable start/stop to work as expected, but still use create to
	 * init early.
	 */
	private boolean initialized;

	/**
	 * Uses defaults.
	 * 
	 * @jmx:managed-constructor
	 * 
	 * @throws MalformedURLException
	 *             Could not create URL from default (propbably a problem with
	 *             overridden properties).
	 */
	public JDKLoggerService() throws MalformedURLException {
		this(DEFAULT_URL, 60);
	}

	/**
	 * @jmx:managed-constructor
	 * 
	 * @param url
	 *            The configuration URL.
	 */
	public JDKLoggerService(final URL url) {
		this(url, 60);
	}

	/**
	 * @jmx:managed-constructor
	 * 
	 * @param url
	 *            The configuration URL.
	 */
	public JDKLoggerService(final String url) throws MalformedURLException {
		this(Strings.toURL(url), 60);
	}

	/**
	 * @jmx:managed-constructor
	 * 
	 * @param url
	 *            The configuration URL.
	 * @param refreshPeriod
	 *            The refreshPeriod in seconds to wait between each check.
	 */
	public JDKLoggerService(final String url, final int refreshPeriod)
			throws MalformedURLException {
		this(Strings.toURL(url), refreshPeriod);
	}

	/**
	 * @jmx:managed-constructor
	 * 
	 * @param url
	 *            The configuration URL.
	 * @param refreshPeriod
	 *            The refreshPeriod in seconds to wait between each check.
	 */
	public JDKLoggerService(final URL url, final int refreshPeriod) {
		this.configURL = url;
		this.refreshPeriod = refreshPeriod;
	}

	/**
	 * Get the refresh period.
	 * 
	 * @jmx:managed-attribute
	 */
	public int getRefreshPeriod() {
		return refreshPeriod;
	}

	/**
	 * Set the refresh period.
	 * 
	 * @jmx:managed-attribute
	 */
	public void setRefreshPeriod(final int refreshPeriod) {
		this.refreshPeriod = refreshPeriod;
	}

	/**
	 * Get the Log4j configuration URL.
	 * 
	 * @jmx:managed-attribute
	 */
	public URL getConfigurationURL() {
		return configURL;
	}

	/**
	 * Set the Log4j configuration URL.
	 * 
	 * @jmx:managed-attribute
	 */
	public void setConfigurationURL(final URL url) {
		this.configURL = url;
	}

	/**
	 * Sets the level for a logger of the give name.
	 * 
	 * <p>
	 * Values are trimmed before used.
	 * 
	 * @jmx:managed-operation
	 * 
	 * @param name
	 *            The name of the logger to change level
	 * @param levelName
	 *            The name of the level to change the logger to.
	 */
	public void setLoggerLevel(final String name, final String levelName) {
		LogManager.getLoggingMXBean().setLoggerLevel(name, levelName);
		log.info("Level set to " + levelName + " for " + name);
	}

	/**
	 * Sets the levels of each logger specified by the given comma seperated
	 * list of logger names.
	 * 
	 * @jmx:managed-operation
	 * 
	 * @see #setLoggerLevel
	 * 
	 * @param list
	 *            A comma seperated list of logger names.
	 * @param levelName
	 *            The name of the level to change the logger to.
	 */
	public void setLoggerLevels(final String list, final String levelName) {
		StringTokenizer stok = new StringTokenizer(list, ",");

		while (stok.hasMoreTokens()) {
			String name = stok.nextToken();
			setLoggerLevel(name, levelName);
		}
	}

	/**
	 * Gets the level of the logger of the give name.
	 * 
	 * @jmx:managed-operation
	 * 
	 * @param name
	 *            The name of the logger to inspect.
	 */
	public String getLoggerLevel(final String name) {

		return LogManager.getLoggingMXBean().getLoggerLevel(name);
	}

	public List<String> getLoggerNames() {
		return LogManager.getLoggingMXBean().getLoggerNames();

	}

	/**
	 * Force the logging system to reconfigure.
	 * 
	 * @jmx:managed-operation
	 */
	public void reconfigure() throws IOException {
		if (timerTask == null)
			throw new IllegalStateException("Service stopped or not started");

		timerTask.reconfigure();
	}

	/**
	 * Hack to reconfigure and change the URL. This is needed until we have a
	 * JMX HTML Adapter that can use PropertyEditor to coerce.
	 * 
	 * @jmx:managed-operation
	 * 
	 * @param url
	 *            The new configuration url
	 */
	public void reconfigure(final String url) throws IOException,
			MalformedURLException {
		setConfigurationURL(Strings.toURL(url));
		reconfigure();
	}

	// /////////////////////////////////////////////////////////////////////////
	// Concrete Service Overrides //
	// /////////////////////////////////////////////////////////////////////////

	protected ObjectName getObjectName(MBeanServer server, ObjectName name)
			throws MalformedObjectNameException {
		return name == null ? OBJECT_NAME : name;
	}

	private void setup() throws Exception {
		if (initialized)
			return;

		timerTask = new URLWatchTimerTask();
		timerTask.run();
		timer.schedule(timerTask, 1000 * refreshPeriod, 1000 * refreshPeriod);

		initialized = true;
	}

	protected void createService() throws Exception {
		setup();
	}

	protected void startService() throws Exception {
		setup();
	}

	protected void stopService() throws Exception {
		timerTask.cancel();
		timerTask = null;

		// allow start to re-initalize
		initialized = false;
	}

	// /////////////////////////////////////////////////////////////////////////
	// URL Watching Timer Task //
	// /////////////////////////////////////////////////////////////////////////

	/**
	 * A timer task to check when a URL changes (based on last modified time)
	 * and reconfigure Log4j.
	 */
	private class URLWatchTimerTask extends TimerTask {
		// private Logger log = Logger
		// .getLogger(URLWatchTimerTask.class.getName());

		private long lastConfigured = -1;

		public void run() {
			log.trace("Checking if configuration changed");

			boolean trace = log.isTraceEnabled();

			try {
				URLConnection conn = configURL.openConnection();
				if (trace)
					log.trace("connection: " + conn);

				long lastModified = conn.getLastModified();
				if (trace) {
					log.trace("last configured: " + lastConfigured);
					log.trace("last modified: " + lastModified);
				}

				if (lastConfigured < lastModified) {
					reconfigure(conn);
				}
			} catch (Exception e) {
				log.warn("Failed to check URL: " + configURL, e);
			}
		}

		public void reconfigure() throws IOException {
			URLConnection conn = configURL.openConnection();
			reconfigure(conn);
		}

		private void reconfigure(final URLConnection conn) {
			log.info("Configuring from URL: " + configURL);

			boolean trace = log.isTraceEnabled();

			String filename = configURL.getFile().toLowerCase();
			try {
				InputStream st = conn.getInputStream();
				// Dump our config if trace is enabled
				if (trace) {
					log.trace("filename: " + filename);
					Streams.copy(st, System.out);
				}

				LogManager.getLogManager().readConfiguration(st);
				st.close();
			} catch (Exception e) {
				log.error("Failed to read config", e);
			}

			// and then rember when we were last changed
			lastConfigured = System.currentTimeMillis();
		}
	}
}
